#include "level.h"

#include <fstream>
#include <iostream>


#include "player.h"
#include "entity.h"

#include "utils.h"

#define NB_TILES_W 128
#define NB_TILES_H 72

void createLevel(Level* lvl, int width, int height) {
      lvl->width = width;
      lvl->height = height;

      int tile_width = (float)width / NB_TILES_W;
      int tile_height = (float)height / NB_TILES_H;

      lvl->gridVertices.resize(NB_TILES_W * NB_TILES_H);
      for (int j = 0; j < NB_TILES_H; j++) {
            for (int i = 0; i < NB_TILES_W; i++) {
                  sf::Vertex v;
                  v.position = sf::Vector2f(i * tile_width, j * tile_height);
                  lvl->gridVertices.push_back(v);
            }
      }
}

int snapToGrid(int nbTiles, int gridSize, int value) {
      float tileSize = gridSize / nbTiles;
      return value % int(tileSize) < tileSize / 2.0f ?
                                     std::floor(value / tileSize) * tileSize :
            std::ceil(value / tileSize) * tileSize;
}

void addSegmentToLevel(Level* lvl, sf::Vector2f start, sf::Vector2f end, LineType type) {
      sf::Vector2f snappedStart = sf::Vector2f(
                                               snapToGrid(NB_TILES_W, lvl->width, start.x),
                                               snapToGrid(NB_TILES_H, lvl->height, start.y));
      sf::Vector2f snappedEnd = sf::Vector2f(
                                             snapToGrid(NB_TILES_W, lvl->width, end.x),
                                             snapToGrid(NB_TILES_H, lvl->height, end.y));

      if (snappedStart == snappedEnd)
            return;

      lvl->segments.push_back({ snappedStart, snappedEnd, type });
}


void drawLevel(Level* lvl, sf::RenderWindow& window) {

      for (const auto& line : lvl->segments) {
            sf::Color c = line.type == LineType::BAD ? sf::Color(255, 0, 0, 100) : COLOR_LEVEL;
            sf::Vertex l[] =
                  {
			sf::Vertex(line.a, c),
			sf::Vertex(line.b, c)
                  };
            window.draw(l, 2, sf::Lines);
      }

      // draw grid
    #if DISPLAY_GRID
      window.draw(&lvl->gridVertices[0], lvl->gridVertices.size(), sf::Points);
    #endif
}

void saveLevel(Level* lvl, std::string filename, std::vector<Entity>* entities, sf::Vector2f* p1Pos, sf::Vector2f* p2Pos) {
	std::ofstream myfile(filename.c_str());
	if (myfile.is_open())
	{
		myfile << "1 " << (int)p1Pos->x << " " << (int)p1Pos->y << std::endl;
		myfile << "2 " << (int)p2Pos->x << " " << (int)p2Pos->y << std::endl;
		for (const auto& entity : *entities) {
			myfile << "E " << (int)entity.radius << " " << (int)entity.value << " " << (int)entity.moveInfo.pos.x << " " << (int)entity.moveInfo.pos.y << std::endl;
		}
		for (const auto& line : lvl->segments) {
			myfile << "S " << line.a.x << " " << line.a.y << " " << line.b.x << " " << line.b.y << " " << line.type << std::endl;
		}
		myfile.close();
	}
	else std::cerr << "Unable to open file";
}

void loadLevel(Level* lvl, std::string filename, std::vector<Entity>* entities, sf::Vector2f* p1Pos, sf::Vector2f* p2Pos) {
    lvl->segments.clear();
    entities->clear();
	std::ifstream myfile(filename.c_str());
	if (myfile.is_open())
	{
		char type;
		while (myfile.good()) {
			myfile >> type;

			if (type == '1') {
				myfile >> p1Pos->x >> p1Pos->y;
            }
			else if (type == '2') {
				myfile >> p2Pos->x >> p2Pos->y;
            }
			else if (type == 'E') {
				int radius, value, posX, posY;
				myfile >> radius >> value >> posX >> posY;
				Entity e;
				createEntity(&e, sf::Vector2f(posX, posY), radius, sf::Color(200, 200, 200, 200), value);
				entities->push_back(e);
			}
			else if (type == 'S') {
				int a, b, c, d, e;
				myfile >> a >> b >> c >> d >> e;
				Line l;
				l.a.x = a;
				l.a.y = b;
				l.b.x = c;
				l.b.y = d;
				l.type = static_cast<LineType>(e);
				lvl->segments.push_back(l);
			}
		}
		myfile.close();
	}
	else std::cerr << "Unable to open file";
}

// http://stackoverflow.com/questions/563198/how-do-you-detect-where-two-line-segments-intersect#565282
bool getEdgeIntersection(sf::Vector2f v0s0, sf::Vector2f v0s1, sf::Vector2f v1s0,
                         sf::Vector2f v1s1, sf::Vector2f* intersection)
{
      float epsilon = 0.0001;
      sf::Vector2f s1 = v0s1 - v0s0;
      sf::Vector2f s2 = v1s1 - v1s0;

      float s, t, det;
      det = -s2.x * s1.y + s1.x * s2.y;
      if (det < epsilon && det > epsilon) return 0;

      s = (-s1.y * (v0s0.x - v1s0.x) + s1.x * (v0s0.y - v1s0.y)) / det;
      t = (s2.x * (v0s0.y - v1s0.y) - s2.y * (v0s0.x - v1s0.x)) / det;

      if (s >= 0 && s <= 1.0 && t >= 0 && t <= 1.0) {
            intersection->x = v0s0.x + t * s1.x;
            intersection->y = v0s0.y + t * s1.y;
            return 1;
      }
      return 0;
}

bool getEdgeIntersection(sf::Vector2f v0s0, sf::Vector2f v0s1, sf::Vector2f v1s0,
                         sf::Vector2f v1s1, sf::Vector2f* intersection, float* tGlobal)
{
      float epsilon = 0.0001;
      sf::Vector2f s1 = v0s1 - v0s0;
      sf::Vector2f s2 = v1s1 - v1s0;

      float s, t, det;
      det = -s2.x * s1.y + s1.x * s2.y;
      if (det < epsilon && det > epsilon) return 0;

      s = (-s1.y * (v0s0.x - v1s0.x) + s1.x * (v0s0.y - v1s0.y)) / det;
      t = (s2.x * (v0s0.y - v1s0.y) - s2.y * (v0s0.x - v1s0.x)) / det;

      if (s >= 0 && s <= 1.0 && t >= 0 && t <= 1.0) {
            intersection->x = v0s0.x + t * s1.x;
            intersection->y = v0s0.y + t * s1.y;
            *tGlobal = t;
            return 1;
      }
      return 0;
}


bool entityPlayerVisibility(Player* player, Entity* entity, std::vector<Line> segments, float* dist) {
      sf::Vector2f playerPos = player->moveInfo.pos;
      sf::Vector2f entityPos = entity->moveInfo.pos;

      *dist = length(entityPos - playerPos);
      sf::Vector2f intersection;
      for (const auto& line : segments) {
            if (getEdgeIntersection(playerPos, entityPos, line.a, line.b, &intersection)) {
                  return false;
            }
      }
      return true;
}
